# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


__doc__ = (
    "A collection of conversion tables & precomputed values to "
    "standardize, speed up, & clean up their usage in the package."
)


from aiootp._typing import Typing as t

from . import collect_non_private_constants


class Tables:
    """
    Conversion tables that are made available throughout the package.
    """

    __slots__ = ()

    ASCII_95 = "".join(chr(val) for val in range(32, 127))
    ASCII_128 = "".join(chr(val) for val in range(128))
    BASE_36 = "0123456789abcdefghijklmnopqrstuvwxyz"
    BASE_38 = "-0123456789_abcdefghijklmnopqrstuvwxyz"
    BASE_64 = "+/0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    BASE_64 += "abcdefghijklmnopqrstuvwxyz"
    BYTE_TUPLE = tuple(i.to_bytes(1, "big") for i in range(256))
    BYTES = b"".join(BYTE_TUPLE)
    HEX = "0123456789abcdef"
    ONION_CHARS = "234567abcdefghijklmnopqrstuvwxyz"
    URL_SAFE = "-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    URL_SAFE += "_abcdefghijklmnopqrstuvwxyz"


PRIMES = {
    # A static dict of primes cached for efficiency & convenience. The
    # primes are organized such that each integer key represents a bit
    # length, & each of the dict's values is a list of two primes, each
    # the first & last primes of that bit length, respectively.
    2: ["02", "03"],
    3: ["05", "07"],
    4: ["0b", "0d"],
    5: ["11", "1f"],
    6: ["25", "3d"],
    7: ["43", "7f"],
    8: ["83", "fb"],
    9: [2 * "01", "01" + "fd"],
    10: ["02" + "09", "03" + "fd"],
    11: ["04" + "07", "07" + "f7"],
    12: ["08" + "05", "0f" + "fd"],
    13: ["10" + "03", "1f" + "ff"],
    14: ["20" + "11", "3f" + "fd"],
    15: ["40" + "1b", "7f" + "ed"],
    16: ["80" + "03", "ff" + "f1"],
    17: ["01" + "00" + "01", "01" + 2 * "ff"],
    18: ["02" + "00" + "1d", "03" + "ff" + "fb"],
    19: ["04" + "00" + "03", "07" + 2 * "ff"],
    20: ["08" + "00" + "15", "0f" + "ff" + "fd"],
    21: ["10" + "00" + "07", "1f" + "ff" + "f7"],
    22: ["20" + "00" + "11", "3f" + "ff" + "fd"],
    23: ["40" + "00" + "0f", "7f" + "ff" + "f1"],
    24: ["80" + "00" + "09", 2 * "ff" + "fd"],
    25: ["01" + 2 * "00" + "2b", "01" + 2 * "ff" + "d9"],
    26: ["02" + 2 * "00" + "23", "03" + 2 * "ff" + "fb"],
    27: ["04" + 2 * "00" + "0f", "07" + 2 * "ff" + "d9"],
    28: ["08" + 2 * "00" + "1d", "0f" + 2 * "ff" + "c7"],
    29: ["10" + 2 * "00" + "03", "1f" + 2 * "ff" + "fd"],
    30: ["20" + 2 * "00" + "0b", "3f" + 2 * "ff" + "dd"],
    31: ["40" + 2 * "00" + "03", "7f" + 3 * "ff"],
    32: ["80" + 2 * "00" + "0b", 3 * "ff" + "fb"],
    33: ["01" + 3 * "00" + "0f", "01" + 3 * "ff" + "f7"],
    34: ["02" + 3 * "00" + "11", "03" + 3 * "ff" + "d7"],
    35: ["04" + 3 * "00" + "19", "07" + 3 * "ff" + "e1"],
    36: ["08" + 3 * "00" + "35", "0f" + 3 * "ff" + "fb"],
    37: ["10" + 3 * "00" + "1f", "1f" + 3 * "ff" + "e7"],
    38: ["20" + 3 * "00" + "09", "3f" + 3 * "ff" + "d3"],
    39: ["40" + 3 * "00" + "07", "7f" + 3 * "ff" + "f9"],
    40: ["80" + 3 * "00" + "17", 4 * "ff" + "a9"],
    41: ["01" + 4 * "00" + "0f", "01" + 4 * "ff" + "eb"],
    42: ["02" + 4 * "00" + "1b", "03" + 4 * "ff" + "f5"],
    43: ["04" + 4 * "00" + "0f", "07" + 4 * "ff" + "c7"],
    44: ["08" + 4 * "00" + "1d", "0f" + 4 * "ff" + "ef"],
    45: ["10" + 4 * "00" + "07", "1f" + 4 * "ff" + "c9"],
    46: ["20" + 4 * "00" + "3b", "3f" + 4 * "ff" + "eb"],
    47: ["40" + 4 * "00" + "0f", "7f" + 4 * "ff" + "8d"],
    48: ["80" + 4 * "00" + "05", 5 * "ff" + "c5"],
    49: ["01" + 5 * "00" + "15", "01" + 5 * "ff" + "af"],
    50: ["02" + 5 * "00" + "45", "03" + 5 * "ff" + "e5"],
    51: ["04" + 5 * "00" + "37", "07" + 5 * "ff" + "7f"],
    52: ["08" + 5 * "00" + "15", "0f" + 5 * "ff" + "d1"],
    53: ["10" + 5 * "00" + "15", "1f" + 5 * "ff" + "91"],
    54: ["20" + 5 * "00" + "05", "3f" + 5 * "ff" + "df"],
    55: ["40" + 5 * "00" + "9f", "7f" + 5 * "ff" + "c9"],
    56: ["80" + 5 * "00" + "03", 6 * "ff" + "fb"],
    57: ["01" + 6 * "00" + "51", "01" + 6 * "ff" + "f3"],
    58: ["02" + 6 * "00" + "09", "03" + 6 * "ff" + "e5"],
    59: ["04" + 6 * "00" + "45", "07" + 6 * "ff" + "c9"],
    60: ["08" + 6 * "00" + "83", "0f" + 6 * "ff" + "a3"],
    61: ["10" + 6 * "00" + "21", "1f" + 7 * "ff"],
    62: ["20" + 6 * "00" + "0f", "3f" + 6 * "ff" + "c7"],
    63: ["40" + 6 * "00" + "87", "7f" + 6 * "ff" + "e7"],
    64: ["80" + 6 * "00" + "1d", 7 * "ff" + "c5"],
    65: ["01" + 7 * "00" + "0d", "01" + 7 * "ff" + "cf"],
    66: ["02" + 7 * "00" + "83", "03" + 7 * "ff" + "fb"],
    67: ["04" + 7 * "00" + "09", "07" + 7 * "ff" + "ed"],
    68: ["08" + 7 * "00" + "03", "0f" + 7 * "ff" + "e9"],
    69: ["10" + 7 * "00" + "21", "1f" + 7 * "ff" + "ed"],
    70: ["20" + 7 * "00" + "1d", "3f" + 7 * "ff" + "dd"],
    71: ["40" + 7 * "00" + "19", "7f" + 7 * "ff" + "19"],
    72: ["80" + 7 * "00" + "0b", 8 * "ff" + "a3"],
    73: ["01" + 8 * "00" + "0f", "01" + 8 * "ff" + "bb"],
    74: ["02" + 8 * "00" + "1d", "03" + 8 * "ff" + "dd"],
    75: ["04" + 8 * "00" + "25", "07" + 8 * "ff" + "9f"],
    76: ["08" + 8 * "00" + "21", "0f" + 8 * "ff" + "f1"],
    77: ["10" + 8 * "00" + "0f", "1f" + 8 * "ff" + "df"],
    78: ["20" + 8 * "00" + "0b", "3f" + 8 * "ff" + "f5"],
    79: ["40" + 8 * "00" + "07", "7f" + 8 * "ff" + "bd"],
    80: ["80" + 8 * "00" + "17", 9 * "ff" + "bf"],
    81: ["01" + 9 * "00" + "0d", "01" + 9 * "ff" + "cd"],
    82: ["02" + 9 * "00" + "11", "03" + 9 * "ff" + "c7"],
    83: ["04" + 9 * "00" + "09", "07" + 9 * "ff" + "c9"],
    84: ["08" + 9 * "00" + "4b", "0f" + 9 * "ff" + "dd"],
    85: ["10" + 9 * "00" + "03", "1f" + 9 * "ff" + "ed"],
    86: ["20" + 9 * "00" + "ab", "3f" + 9 * "ff" + "dd"],
    87: ["40" + 9 * "00" + "1b", "7f" + 9 * "ff" + "bd"],
    88: ["80" + 9 * "00" + "27", 9 * "ff" + "fe" + "d5"],
    89: ["01" + 10 * "00" + "07", "01" + 11 * "ff"],
    90: ["02" + 10 * "00" + "1d", "03" + 10 * "ff" + "df"],
    91: ["04" + 10 * "00" + "85", "07" + 10 * "ff" + "d3"],
    92: ["08" + 10 * "00" + "3b", "0f" + 10 * "ff" + "ad"],
    93: ["10" + 10 * "00" + "19", "1f" + 10 * "ff" + "e7"],
    94: ["20" + 10 * "00" + "69", "3f" + 10 * "ff" + "fd"],
    95: ["40" + 10 * "00" + "81", "7f" + 10 * "ff" + "f1"],
    96: ["80" + 10 * "00" + "09", 11 * "ff" + "ef"],
    97: ["01" + 11 * "00" + "3d", "01" + 11 * "ff" + "73"],
    98: ["02" + 11 * "00" + "69", "03" + 11 * "ff" + "cd"],
    99: ["04" + 11 * "00" + "07", "07" + 11 * "ff" + "8d"],
    100: ["08" + 11 * "00" + "ff", "0f" + 11 * "ff" + "f1"],
    101: ["10" + 10 * "00" + "01" + "15", "1f" + 11 * "ff" + "bb"],
    102: ["20" + 11 * "00" + "51", "3f" + 11 * "ff" + "df"],
    103: ["40" + 10 * "00" + "01" + "0b", "7f" + 11 * "ff" + "9f"],
    104: ["80" + 11 * "00" + "51", 12 * "ff" + "ef"],
    105: ["01" + 12 * "00" + "6f", "01" + 12 * "ff" + "f3"],
    106: ["02" + 12 * "00" + "27", "03" + 12 * "ff" + "8b"],
    107: ["04" + 12 * "00" + "63", "07" + 13 * "ff"],
    108: ["08" + 12 * "00" + "27", "0f" + 12 * "ff" + "c5"],
    109: ["10" + 12 * "00" + "21", "1f" + 12 * "ff" + "e1"],
    110: ["20" + 12 * "00" + "93", "3f" + 12 * "ff" + "eb"],
    111: ["40" + 12 * "00" + "1b", "7f" + 12 * "ff" + "db"],
    112: ["80" + 12 * "00" + "33", 13 * "ff" + "b5"],
    113: ["01" + 13 * "00" + "19", "01" + 13 * "ff" + "7b"],
    114: ["02" + 12 * "00" + "01" + "19", "03" + 13 * "ff" + "f5"],
    115: ["04" + 13 * "00" + "2b", "07" + 13 * "ff" + "bd"],
    116: ["08" + 13 * "00" + "47", "0f" + 13 * "ff" + "fd"],
    117: ["10" + 13 * "00" + "21", "1f" + 12 * "ff" + "fe" + "e9"],
    118: ["20" + 13 * "00" + "1d", "3f" + 13 * "ff" + "fb"],
    119: ["40" + 13 * "00" + "19", "7f" + 13 * "ff" + "bb"],
    120: ["80" + 13 * "00" + "09", 14 * "ff" + "89"],
    121: ["01" + 13 * "00" + "01" + "c3", "01" + 14 * "ff" + "b7"],
    122: ["02" + 14 * "00" + "29", "03" + 14 * "ff" + "fd"],
    123: ["04" + 13 * "00" + "01" + "15", "07" + 14 * "ff" + "bd"],
    124: ["08" + 14 * "00" + "a5", "0f" + 14 * "ff" + "c5"],
    125: ["10" + 14 * "00" + "43", "1f" + 14 * "ff" + "f7"],
    126: ["20" + 14 * "00" + "1b", "3f" + 14 * "ff" + "77"],
    127: ["40" + 14 * "00" + "07", "7f" + 15 * "ff"],
    128: ["80" + 14 * "00" + "1d", 15 * "ff" + "61"],
    129: ["01" + 15 * "00" + "33", "01" + 15 * "ff" + "e7"],
    130: ["02" + 15 * "00" + "11", "03" + 15 * "ff" + "fb"],
    131: ["04" + 15 * "00" + "a9", "07" + 15 * "ff" + "bb"],
    132: ["08" + 15 * "00" + "27", "0f" + 14 * "ff" + "fe" + "a5"],
    133: ["10" + 15 * "00" + "43", "1f" + 15 * "ff" + "9d"],
    134: ["20" + 15 * "00" + "1b", "3f" + 15 * "ff" + "d3"],
    135: ["40" + 15 * "00" + "1b", "7f" + 15 * "ff" + "d3"],
    136: ["80" + 15 * "00" + "21", 16 * "ff" + "8f"],
    137: ["01" + 16 * "00" + "55", "01" + 16 * "ff" + "f3"],
    138: ["02" + 16 * "00" + "9b", "03" + 16 * "ff" + "97"],
    139: ["04" + 16 * "00" + "57", "07" + 16 * "ff" + "45"],
    140: ["08" + 16 * "00" + "9b", "0f" + 16 * "ff" + "e5"],
    141: ["10" + 16 * "00" + "25", "1f" + 16 * "ff" + "f7"],
    142: ["20" + 16 * "00" + "05", "3f" + 16 * "ff" + "91"],
    143: ["40" + 16 * "00" + "d9", "7f" + 16 * "ff" + "bb"],
    144: ["80" + 16 * "00" + "05", 17 * "ff" + "ad"],
    145: ["01" + 17 * "00" + "af", "01" + 17 * "ff" + "69"],
    146: ["02" + 17 * "00" + "1b", "03" + 17 * "ff" + "67"],
    147: ["04" + 17 * "00" + "55", "07" + 17 * "ff" + "6f"],
    148: ["08" + 17 * "00" + "33", "0f" + 17 * "ff" + "59"],
    149: ["10" + 17 * "00" + "5b", "1f" + 17 * "ff" + "e1"],
    150: ["20" + 17 * "00" + "45", "3f" + 17 * "ff" + "fd"],
    151: ["40" + 17 * "00" + "93", "7f" + 17 * "ff" + "3d"],
    152: ["80" + 17 * "00" + "2d", 18 * "ff" + "ef"],
    153: ["01" + 18 * "00" + "fd", "01" + 18 * "ff" + "bb"],
    154: ["02" + 18 * "00" + "5f", "03" + 18 * "ff" + "0d"],
    155: ["04" + 18 * "00" + "1b", "07" + 18 * "ff" + "e1"],
    156: ["08" + 18 * "00" + "0f", "0f" + 18 * "ff" + "71"],
    157: ["10" + 18 * "00" + "2d", "1f" + 18 * "ff" + "ed"],
    158: ["20" + 18 * "00" + "45", "3f" + 18 * "ff" + "f1"],
    159: ["40" + 18 * "00" + "61", "7f" + 18 * "ff" + "a5"],
    160: ["80" + 17 * "00" + "01" + "2b", 19 * "ff" + "d1"],
    161: ["01" + 19 * "00" + "07", "01" + 19 * "ff" + "61"],
    162: ["02" + 19 * "00" + "6b", "03" + 19 * "ff" + "9b"],
    163: ["04" + 19 * "00" + "13", "07" + 19 * "ff" + "c9"],
    164: ["08" + 19 * "00" + "15", "0f" + 19 * "ff" + "c1"],
    165: ["10" + 19 * "00" + "75", "1f" + 19 * "ff" + "e7"],
    166: ["20" + 19 * "00" + "8d", "3f" + 19 * "ff" + "fb"],
    167: ["40" + 19 * "00" + "55", "7f" + 19 * "ff" + "79"],
    168: ["80" + 19 * "00" + "53", 19 * "ff" + "fe" + "ff"],
    169: ["01" + 20 * "00" + "57", "01" + 19 * "ff" + "fd" + "7d"],
    170: ["02" + 20 * "00" + "93", "03" + 20 * "ff" + "71"],
    171: ["04" + 20 * "00" + "31", "07" + 20 * "ff" + "ed"],
    172: ["08" + 20 * "00" + "81", "0f" + 20 * "ff" + "a1"],
    173: ["10" + 20 * "00" + "69", "1f" + 20 * "ff" + "c9"],
    174: ["20" + 20 * "00" + "4d", "3f" + 20 * "ff" + "fd"],
    175: ["40" + 20 * "00" + "07", "7f" + 20 * "ff" + "1b"],
    176: ["80" + 20 * "00" + "09", 21 * "ff" + "17"],
    177: ["01" + 20 * "00" + "01" + "ab", "01" + 20 * "ff" + "fe" + "ad"],
    178: ["02" + 21 * "00" + "4b", "03" + 21 * "ff" + "d7"],
    179: ["04" + 21 * "00" + "57", "07" + 21 * "ff" + "cf"],
    180: ["08" + 20 * "00" + "01" + "35", "0f" + 21 * "ff" + "d1"],
    181: ["10" + 21 * "00" + "0f", "1f" + 21 * "ff" + "5b"],
    182: ["20" + 21 * "00" + "a5", "3f" + 21 * "ff" + "5f"],
    183: ["40" + 21 * "00" + "31", "7f" + 21 * "ff" + "6d"],
    184: ["80" + 21 * "00" + "d7", 22 * "ff" + "df"],
    185: ["01" + 22 * "00" + "1b", "01" + 21 * "ff" + "fe" + "d1"],
    186: ["02" + 22 * "00" + "9f", "03" + 21 * "ff" + "fe" + "8d"],
    187: ["04" + 22 * "00" + "cd", "07" + 22 * "ff" + "ab"],
    188: ["08" + 21 * "00" + "01" + "2f", "0f" + 22 * "ff" + "83"],
    189: ["10" + 22 * "00" + "39", "1f" + 22 * "ff" + "e7"],
    190: ["20" + 22 * "00" + "23", "3f" + 22 * "ff" + "f5"],
    191: ["40" + 22 * "00" + "81", "7f" + 22 * "ff" + "ed"],
    192: ["80" + 22 * "00" + "05", 23 * "ff" + "13"],
    193: ["01" + 23 * "00" + "85", "01" + 23 * "ff" + "e1"],
    194: ["02" + 23 * "00" + "41", "03" + 23 * "ff" + "df"],
    195: ["04" + 23 * "00" + "1b", "07" + 23 * "ff" + "79"],
    196: ["08" + 23 * "00" + "23", "0f" + 23 * "ff" + "f1"],
    197: ["10" + 23 * "00" + "15", "1f" + 23 * "ff" + "b5"],
    198: ["20" + 23 * "00" + "6b", "3f" + 23 * "ff" + "ef"],
    199: ["40" + 23 * "00" + "0f", "7f" + 23 * "ff" + "cf"],
    200: ["80" + 23 * "00" + "65", 24 * "ff" + "b5"],
    201: ["01" + 24 * "00" + "eb", "01" + 24 * "ff" + "c9"],
    202: ["02" + 23 * "00" + "01" + "5f", "03" + 24 * "ff" + "49"],
    203: ["04" + 24 * "00" + "43", "07" + 24 * "ff" + "61"],
    204: ["08" + 24 * "00" + "0f", "0f" + 24 * "ff" + "59"],
    205: ["10" + 24 * "00" + "07", "1f" + 24 * "ff" + "af"],
    206: ["20" + 23 * "00" + "02" + "45", "3f" + 24 * "ff" + "fb"],
    207: ["40" + 24 * "00" + "21", "7f" + 24 * "ff" + "a5"],
    208: ["80" + 24 * "00" + "cb", 24 * "ff" + "fe" + "d5"],
    209: ["01" + 24 * "00" + "01" + "77", "01" + 25 * "ff" + "df"],
    210: ["02" + 25 * "00" + "2f", "03" + 25 * "ff" + "d1"],
    211: ["04" + 25 * "00" + "21", "07" + 25 * "ff" + "51"],
    212: ["08" + 25 * "00" + "47", "0f" + 25 * "ff" + "e9"],
    213: ["10" + 25 * "00" + "39", "1f" + 25 * "ff" + "fd"],
    214: ["20" + 25 * "00" + "4b", "3f" + 25 * "ff" + "47"],
    215: ["40" + 25 * "00" + "07", "7f" + 25 * "ff" + "63"],
    216: ["80" + 25 * "00" + "fb", 25 * "ff" + "fe" + "87"],
    217: ["01" + 25 * "00" + "01" + "a7", "01" + 26 * "ff" + "c3"],
    218: ["02" + 26 * "00" + "81", "03" + 26 * "ff" + "df"],
    219: ["04" + 26 * "00" + "a3", "07" + 26 * "ff" + "87"],
    220: ["08" + 26 * "00" + "b9", "0f" + 26 * "ff" + "b3"],
    221: ["10" + 26 * "00" + "d9", "1f" + 26 * "ff" + "fd"],
    222: ["20" + 26 * "00" + "51", "3f" + 26 * "ff" + "8b"],
    223: ["40" + 26 * "00" + "31", "7f" + 26 * "ff" + "15"],
    224: ["80" + 26 * "00" + "bd", 27 * "ff" + "c1"],
    225: ["01" + 26 * "00" + "02" + "df", "01" + 27 * "ff" + "cf"],
    226: ["02" + 27 * "00" + "77", "03" + 27 * "ff" + "fb"],
    227: ["04" + 26 * "00" + "02" + "df", "07" + 26 * "ff" + "fe" + "6b"],
    228: ["08" + 26 * "00" + "01" + "e3", "0f" + 27 * "ff" + "a3"],
    229: ["10" + 27 * "00" + "03", "1f" + 27 * "ff" + "a5"],
    230: ["20" + 27 * "00" + "f9", "3f" + 27 * "ff" + "e5"],
    231: ["40" + 27 * "00" + "43", "7f" + 27 * "ff" + "5b"],
    232: ["80" + 27 * "00" + "69", 27 * "ff" + "fd" + "c9"],
    233: ["01" + 27 * "00" + "01" + "65", "01" + 28 * "ff" + "fd"],
    234: ["02" + 27 * "00" + "01" + "af", "03" + 28 * "ff" + "ad"],
    235: ["04" + 28 * "00" + "2b", "07" + 28 * "ff" + "f1"],
    236: ["08" + 28 * "00" + "51", "0f" + 28 * "ff" + "2f"],
    237: ["10" + 28 * "00" + "19", "1f" + 28 * "ff" + "4b"],
    238: ["20" + 28 * "00" + "f9", "3f" + 28 * "ff" + "5f"],
    239: ["40" + 28 * "00" + "43", "7f" + 28 * "ff" + "a9"],
    240: ["80" + 28 * "00" + "1d", 28 * "ff" + "fe" + "2d"],
    241: ["01" + 29 * "00" + "73", "01" + 29 * "ff" + "d9"],
    242: ["02" + 28 * "00" + "01" + "05", "03" + 29 * "ff" + "c1"],
    243: ["04" + 29 * "00" + "45", "07" + 29 * "ff" + "f7"],
    244: ["08" + 29 * "00" + "3b", "0f" + 29 * "ff" + "43"],
    245: ["10" + 29 * "00" + "85", "1f" + 29 * "ff" + "5d"],
    246: ["20" + 28 * "00" + "01" + "3b", "3f" + 29 * "ff" + "95"],
    247: ["40" + 28 * "00" + "01" + "51", "7f" + 29 * "ff" + "af"],
    248: ["80" + 29 * "00" + "3f", 30 * "ff" + "13"],
    249: ["01" + 30 * "00" + "51", "01" + 30 * "ff" + "b5"],
    250: ["02" + 30 * "00" + "77", "03" + 30 * "ff" + "31"],
    251: ["04" + 30 * "00" + "19", "07" + 30 * "ff" + "f7"],
    252: ["08" + 30 * "00" + "41", "0f" + 30 * "ff" + "7f"],
    253: ["10" + 29 * "00" + "01" + "a5", "1f" + 29 * "ff" + "fe" + "ef"],
    254: ["20" + 30 * "00" + "27", "3f" + 30 * "ff" + "0b"],
    255: ["40" + 30 * "00" + "4f", "7f" + 30 * "ff" + "ed"],
    256: ["80" + 30 * "00" + "5f", 31 * "ff" + "43"],
    257: ["01" + 30 * "00" + "01" + "29", "01" + 31 * "ff" + "a3"],
    383: ["40" + 46 * "00" + "ff", "7f" + 46 * "ff" + "e1"],
    384: ["80" + 45 * "00" + "01" + "71", 46 * "ff" + "fe" + "c3"],
    385: ["01" + 47 * "00" + "e7", "01" + 46 * "ff" + "fe" + "f7"],
    511: ["40" + 62 * "00" + "0f", "7f" + 62 * "ff" + "45"],
    512: ["80" + 62 * "00" + "6f", 62 * "ff" + "fd" + "c7"],
    513: ["01" + 63 * "00" + "4b", "01" + 62 * "ff" + "fe" + "43"],
    1023: [
        "40" + 125 * "00" + "05" + "a3",
        "7f" + 125 * "ff" + "fe" + "97",
    ],
    1024: ["80" + 125 * "00" + "04" + "83", 127 * "ff" + "97"],
    1025: [
        "01" + 126 * "00" + "02" + "83",
        "01" + 126 * "ff" + "f5" + "8f",
    ],
    1535: [
        "40" + 189 * "00" + "01" + "09",
        "7f" + 189 * "ff" + "fa" + "db",
    ],
    1536: ["80" + 189 * "00" + "02" + "bb", 190 * "ff" + "f2" + "83"],
    1537: ["01" + 191 * "00" + "4b", "01" + 191 * "ff" + "51"],
    2047: ["40" + 253 * "00" + "10" + "33", "7f" + 254 * "ff" + "ab"],
    2048: ["80" + 253 * "00" + "07" + "7f", 254 * "ff" + "f9" + "eb"],
    2049: [
        "01" + 254 * "00" + "03" + "d5",
        "01" + 254 * "ff" + "fa" + "57",
    ],
    2559: [
        "40" + 317 * "00" + "03" + "87",
        "7f" + 317 * "ff" + "fd" + "3b",
    ],
    2560: ["80" + 317 * "00" + "08" + "eb", 319 * "ff" + "b5"],
    2561: [
        "01" + 318 * "00" + "03" + "87",
        "01" + 318 * "ff" + "fd" + "99",
    ],
    3071: ["40" + 382 * "00" + "c3", "7f" + 382 * "ff" + "45"],
    3072: ["80" + 381 * "00" + "08" + "f3", 383 * "ff" + "d1"],
    3073: [
        "01" + 382 * "00" + "03" + "2d",
        "01" + 382 * "ff" + "fc" + "b7",
    ],
    3583: [
        "40" + 445 * "00" + "09" + "27",
        "7f" + 445 * "ff" + "f3" + "e7",
    ],
    3584: ["80" + 445 * "00" + "10" + "a3", 446 * "ff" + "fe" + "53"],
    3585: ["01" + 447 * "00" + "15", "01" + 446 * "ff" + "fe" + "65"],
    4095: [
        "40" + 509 * "00" + "01" + "e9",
        "7f" + 509 * "ff" + "f9" + "b1",
    ],
    4096: ["80" + 509 * "00" + "02" + "43", 510 * "ff" + "f6" + "0b"],
    4097: [
        "01" + 510 * "00" + "06" + "e1",
        "01" + 510 * "ff" + "fc" + "f1",
    ],
}


for level in PRIMES.values():
    level[0] = int(level[0], 16)
    level[1] = int(level[1], 16)


WORD_LIST = [
    b"abandon",
    b"ability",
    b"able",
    b"about",
    b"above",
    b"absent",
    b"absorb",
    b"abstract",
    b"absurd",
    b"abuse",
    b"access",
    b"accident",
    b"account",
    b"accuse",
    b"achieve",
    b"acid",
    b"acoustic",
    b"acquire",
    b"across",
    b"act",
    b"action",
    b"actor",
    b"actress",
    b"actual",
    b"adapt",
    b"add",
    b"addict",
    b"address",
    b"adjust",
    b"admit",
    b"adult",
    b"advance",
    b"advice",
    b"aerobic",
    b"affair",
    b"afford",
    b"afraid",
    b"again",
    b"age",
    b"agent",
    b"agree",
    b"ahead",
    b"aim",
    b"air",
    b"airport",
    b"aisle",
    b"alarm",
    b"album",
    b"alcohol",
    b"alert",
    b"alien",
    b"all",
    b"alley",
    b"allow",
    b"almost",
    b"alone",
    b"alpha",
    b"already",
    b"also",
    b"alter",
    b"always",
    b"amateur",
    b"amazing",
    b"among",
    b"amount",
    b"amused",
    b"analyst",
    b"anchor",
    b"ancient",
    b"anger",
    b"angle",
    b"angry",
    b"animal",
    b"ankle",
    b"announce",
    b"annual",
    b"another",
    b"answer",
    b"antenna",
    b"antique",
    b"anxiety",
    b"any",
    b"apart",
    b"apology",
    b"appear",
    b"apple",
    b"approve",
    b"april",
    b"arch",
    b"arctic",
    b"area",
    b"arena",
    b"argue",
    b"arm",
    b"armed",
    b"armor",
    b"army",
    b"around",
    b"arrange",
    b"arrest",
    b"arrive",
    b"arrow",
    b"art",
    b"artefact",
    b"artist",
    b"artwork",
    b"ask",
    b"aspect",
    b"assault",
    b"asset",
    b"assist",
    b"assume",
    b"asthma",
    b"athlete",
    b"atom",
    b"attack",
    b"attend",
    b"attitude",
    b"attract",
    b"auction",
    b"audit",
    b"august",
    b"aunt",
    b"author",
    b"auto",
    b"autumn",
    b"average",
    b"avocado",
    b"avoid",
    b"awake",
    b"aware",
    b"away",
    b"awesome",
    b"awful",
    b"awkward",
    b"axis",
    b"baby",
    b"bachelor",
    b"bacon",
    b"badge",
    b"bag",
    b"balance",
    b"balcony",
    b"ball",
    b"bamboo",
    b"banana",
    b"banner",
    b"bar",
    b"barely",
    b"bargain",
    b"barrel",
    b"base",
    b"basic",
    b"basket",
    b"battle",
    b"beach",
    b"bean",
    b"beauty",
    b"because",
    b"become",
    b"beef",
    b"before",
    b"begin",
    b"behave",
    b"behind",
    b"believe",
    b"below",
    b"belt",
    b"bench",
    b"benefit",
    b"best",
    b"betray",
    b"better",
    b"between",
    b"beyond",
    b"bicycle",
    b"bid",
    b"bike",
    b"bind",
    b"biology",
    b"bird",
    b"birth",
    b"bitter",
    b"black",
    b"blade",
    b"blame",
    b"blanket",
    b"blast",
    b"bleak",
    b"bless",
    b"blind",
    b"blood",
    b"blossom",
    b"blouse",
    b"blue",
    b"blur",
    b"blush",
    b"board",
    b"boat",
    b"body",
    b"boil",
    b"bomb",
    b"bone",
    b"bonus",
    b"book",
    b"boost",
    b"border",
    b"boring",
    b"borrow",
    b"boss",
    b"bottom",
    b"bounce",
    b"box",
    b"boy",
    b"bracket",
    b"brain",
    b"brand",
    b"brass",
    b"brave",
    b"bread",
    b"breeze",
    b"brick",
    b"bridge",
    b"brief",
    b"bright",
    b"bring",
    b"brisk",
    b"broccoli",
    b"broken",
    b"bronze",
    b"broom",
    b"brother",
    b"brown",
    b"brush",
    b"bubble",
    b"buddy",
    b"budget",
    b"buffalo",
    b"build",
    b"bulb",
    b"bulk",
    b"bullet",
    b"bundle",
    b"bunker",
    b"burden",
    b"burger",
    b"burst",
    b"bus",
    b"business",
    b"busy",
    b"butter",
    b"buyer",
    b"buzz",
    b"cabbage",
    b"cabin",
    b"cable",
    b"cactus",
    b"cage",
    b"cake",
    b"call",
    b"calm",
    b"camera",
    b"camp",
    b"can",
    b"canal",
    b"cancel",
    b"candy",
    b"cannon",
    b"canoe",
    b"canvas",
    b"canyon",
    b"capable",
    b"capital",
    b"captain",
    b"car",
    b"carbon",
    b"card",
    b"cargo",
    b"carpet",
    b"carry",
    b"cart",
    b"case",
    b"cash",
    b"casino",
    b"castle",
    b"casual",
    b"cat",
    b"catalog",
    b"catch",
    b"category",
    b"cattle",
    b"caught",
    b"cause",
    b"caution",
    b"cave",
    b"ceiling",
    b"celery",
    b"cement",
    b"census",
    b"century",
    b"cereal",
    b"certain",
    b"chair",
    b"chalk",
    b"champion",
    b"change",
    b"chaos",
    b"chapter",
    b"charge",
    b"chase",
    b"chat",
    b"cheap",
    b"check",
    b"cheese",
    b"chef",
    b"cherry",
    b"chest",
    b"chicken",
    b"chief",
    b"child",
    b"chimney",
    b"choice",
    b"choose",
    b"chronic",
    b"chuckle",
    b"chunk",
    b"churn",
    b"cigar",
    b"cinnamon",
    b"circle",
    b"citizen",
    b"city",
    b"civil",
    b"claim",
    b"clap",
    b"clarify",
    b"claw",
    b"clay",
    b"clean",
    b"clerk",
    b"clever",
    b"click",
    b"client",
    b"cliff",
    b"climb",
    b"clinic",
    b"clip",
    b"clock",
    b"clog",
    b"close",
    b"cloth",
    b"cloud",
    b"clown",
    b"club",
    b"clump",
    b"cluster",
    b"clutch",
    b"coach",
    b"coast",
    b"coconut",
    b"code",
    b"coffee",
    b"coil",
    b"coin",
    b"collect",
    b"color",
    b"column",
    b"combine",
    b"come",
    b"comfort",
    b"comic",
    b"common",
    b"company",
    b"concert",
    b"conduct",
    b"confirm",
    b"congress",
    b"connect",
    b"consider",
    b"control",
    b"convince",
    b"cook",
    b"cool",
    b"copper",
    b"copy",
    b"coral",
    b"core",
    b"corn",
    b"correct",
    b"cost",
    b"cotton",
    b"couch",
    b"country",
    b"couple",
    b"course",
    b"cousin",
    b"cover",
    b"coyote",
    b"crack",
    b"cradle",
    b"craft",
    b"cram",
    b"crane",
    b"crash",
    b"crater",
    b"crawl",
    b"crazy",
    b"cream",
    b"credit",
    b"creek",
    b"crew",
    b"cricket",
    b"crime",
    b"crisp",
    b"critic",
    b"crop",
    b"cross",
    b"crouch",
    b"crowd",
    b"crucial",
    b"cruel",
    b"cruise",
    b"crumble",
    b"crunch",
    b"crush",
    b"cry",
    b"crystal",
    b"cube",
    b"culture",
    b"cup",
    b"cupboard",
    b"curious",
    b"current",
    b"curtain",
    b"curve",
    b"cushion",
    b"custom",
    b"cute",
    b"cycle",
    b"dad",
    b"damage",
    b"damp",
    b"dance",
    b"danger",
    b"daring",
    b"dash",
    b"daughter",
    b"dawn",
    b"day",
    b"deal",
    b"debate",
    b"debris",
    b"decade",
    b"december",
    b"decide",
    b"decline",
    b"decorate",
    b"decrease",
    b"deer",
    b"defense",
    b"define",
    b"defy",
    b"degree",
    b"delay",
    b"deliver",
    b"demand",
    b"demise",
    b"denial",
    b"dentist",
    b"deny",
    b"depart",
    b"depend",
    b"deposit",
    b"depth",
    b"deputy",
    b"derive",
    b"describe",
    b"desert",
    b"design",
    b"desk",
    b"despair",
    b"destroy",
    b"detail",
    b"detect",
    b"develop",
    b"device",
    b"devote",
    b"diagram",
    b"dial",
    b"diamond",
    b"diary",
    b"dice",
    b"diesel",
    b"diet",
    b"differ",
    b"digital",
    b"dignity",
    b"dilemma",
    b"dinner",
    b"dinosaur",
    b"direct",
    b"dirt",
    b"disagree",
    b"discover",
    b"disease",
    b"dish",
    b"dismiss",
    b"disorder",
    b"display",
    b"distance",
    b"divert",
    b"divide",
    b"divorce",
    b"dizzy",
    b"doctor",
    b"document",
    b"dog",
    b"doll",
    b"dolphin",
    b"domain",
    b"donate",
    b"donkey",
    b"donor",
    b"door",
    b"dose",
    b"double",
    b"dove",
    b"draft",
    b"dragon",
    b"drama",
    b"drastic",
    b"draw",
    b"dream",
    b"dress",
    b"drift",
    b"drill",
    b"drink",
    b"drip",
    b"drive",
    b"drop",
    b"drum",
    b"dry",
    b"duck",
    b"dumb",
    b"dune",
    b"during",
    b"dust",
    b"dutch",
    b"duty",
    b"dwarf",
    b"dynamic",
    b"eager",
    b"eagle",
    b"early",
    b"earn",
    b"earth",
    b"easily",
    b"east",
    b"easy",
    b"echo",
    b"ecology",
    b"economy",
    b"edge",
    b"edit",
    b"educate",
    b"effort",
    b"egg",
    b"eight",
    b"either",
    b"elbow",
    b"elder",
    b"electric",
    b"elegant",
    b"element",
    b"elephant",
    b"elevator",
    b"elite",
    b"else",
    b"embark",
    b"embody",
    b"embrace",
    b"emerge",
    b"emotion",
    b"employ",
    b"empower",
    b"empty",
    b"enable",
    b"enact",
    b"end",
    b"endless",
    b"endorse",
    b"enemy",
    b"energy",
    b"enforce",
    b"engage",
    b"engine",
    b"enhance",
    b"enjoy",
    b"enlist",
    b"enough",
    b"enrich",
    b"enroll",
    b"ensure",
    b"enter",
    b"entire",
    b"entry",
    b"envelope",
    b"episode",
    b"equal",
    b"equip",
    b"era",
    b"erase",
    b"erode",
    b"erosion",
    b"error",
    b"erupt",
    b"escape",
    b"essay",
    b"essence",
    b"estate",
    b"eternal",
    b"ethics",
    b"evidence",
    b"evil",
    b"evoke",
    b"evolve",
    b"exact",
    b"example",
    b"excess",
    b"exchange",
    b"excite",
    b"exclude",
    b"excuse",
    b"execute",
    b"exercise",
    b"exhaust",
    b"exhibit",
    b"exile",
    b"exist",
    b"exit",
    b"exotic",
    b"expand",
    b"expect",
    b"expire",
    b"explain",
    b"expose",
    b"express",
    b"extend",
    b"extra",
    b"eye",
    b"eyebrow",
    b"fabric",
    b"face",
    b"faculty",
    b"fade",
    b"faint",
    b"faith",
    b"fall",
    b"false",
    b"fame",
    b"family",
    b"famous",
    b"fan",
    b"fancy",
    b"fantasy",
    b"farm",
    b"fashion",
    b"fat",
    b"fatal",
    b"father",
    b"fatigue",
    b"fault",
    b"favorite",
    b"feature",
    b"february",
    b"federal",
    b"fee",
    b"feed",
    b"feel",
    b"female",
    b"fence",
    b"festival",
    b"fetch",
    b"fever",
    b"few",
    b"fiber",
    b"fiction",
    b"field",
    b"figure",
    b"file",
    b"film",
    b"filter",
    b"final",
    b"find",
    b"fine",
    b"finger",
    b"finish",
    b"fire",
    b"firm",
    b"first",
    b"fiscal",
    b"fish",
    b"fit",
    b"fitness",
    b"fix",
    b"flag",
    b"flame",
    b"flash",
    b"flat",
    b"flavor",
    b"flee",
    b"flight",
    b"flip",
    b"float",
    b"flock",
    b"floor",
    b"flower",
    b"fluid",
    b"flush",
    b"fly",
    b"foam",
    b"focus",
    b"fog",
    b"foil",
    b"fold",
    b"follow",
    b"food",
    b"foot",
    b"force",
    b"forest",
    b"forget",
    b"fork",
    b"fortune",
    b"forum",
    b"forward",
    b"fossil",
    b"foster",
    b"found",
    b"fox",
    b"fragile",
    b"frame",
    b"frequent",
    b"fresh",
    b"friend",
    b"fringe",
    b"frog",
    b"front",
    b"frost",
    b"frown",
    b"frozen",
    b"fruit",
    b"fuel",
    b"fun",
    b"funny",
    b"furnace",
    b"fury",
    b"future",
    b"gadget",
    b"gain",
    b"galaxy",
    b"gallery",
    b"game",
    b"gap",
    b"garage",
    b"garbage",
    b"garden",
    b"garlic",
    b"garment",
    b"gas",
    b"gasp",
    b"gate",
    b"gather",
    b"gauge",
    b"gaze",
    b"general",
    b"genius",
    b"genre",
    b"gentle",
    b"genuine",
    b"gesture",
    b"ghost",
    b"giant",
    b"gift",
    b"giggle",
    b"ginger",
    b"giraffe",
    b"girl",
    b"give",
    b"glad",
    b"glance",
    b"glare",
    b"glass",
    b"glide",
    b"glimpse",
    b"globe",
    b"gloom",
    b"glory",
    b"glove",
    b"glow",
    b"glue",
    b"goat",
    b"goddess",
    b"gold",
    b"good",
    b"goose",
    b"gorilla",
    b"gospel",
    b"gossip",
    b"govern",
    b"gown",
    b"grab",
    b"grace",
    b"grain",
    b"grant",
    b"grape",
    b"grass",
    b"gravity",
    b"great",
    b"green",
    b"grid",
    b"grief",
    b"grit",
    b"grocery",
    b"group",
    b"grow",
    b"grunt",
    b"guard",
    b"guess",
    b"guide",
    b"guilt",
    b"guitar",
    b"gun",
    b"gym",
    b"habit",
    b"hair",
    b"half",
    b"hammer",
    b"hamster",
    b"hand",
    b"happy",
    b"harbor",
    b"hard",
    b"harsh",
    b"harvest",
    b"hat",
    b"have",
    b"hawk",
    b"hazard",
    b"head",
    b"health",
    b"heart",
    b"heavy",
    b"hedgehog",
    b"height",
    b"hello",
    b"helmet",
    b"help",
    b"hen",
    b"hero",
    b"hidden",
    b"high",
    b"hill",
    b"hint",
    b"hip",
    b"hire",
    b"history",
    b"hobby",
    b"hockey",
    b"hold",
    b"hole",
    b"holiday",
    b"hollow",
    b"home",
    b"honey",
    b"hood",
    b"hope",
    b"horn",
    b"horror",
    b"horse",
    b"hospital",
    b"host",
    b"hotel",
    b"hour",
    b"hover",
    b"hub",
    b"huge",
    b"human",
    b"humble",
    b"humor",
    b"hundred",
    b"hungry",
    b"hunt",
    b"hurdle",
    b"hurry",
    b"hurt",
    b"husband",
    b"hybrid",
    b"ice",
    b"icon",
    b"idea",
    b"identify",
    b"idle",
    b"ignore",
    b"ill",
    b"illegal",
    b"illness",
    b"image",
    b"imitate",
    b"immense",
    b"immune",
    b"impact",
    b"impose",
    b"improve",
    b"impulse",
    b"inch",
    b"include",
    b"income",
    b"increase",
    b"index",
    b"indicate",
    b"indoor",
    b"industry",
    b"infant",
    b"inflict",
    b"inform",
    b"inhale",
    b"inherit",
    b"initial",
    b"inject",
    b"injury",
    b"inmate",
    b"inner",
    b"innocent",
    b"input",
    b"inquiry",
    b"insane",
    b"insect",
    b"inside",
    b"inspire",
    b"install",
    b"intact",
    b"interest",
    b"into",
    b"invest",
    b"invite",
    b"involve",
    b"iron",
    b"island",
    b"isolate",
    b"issue",
    b"item",
    b"ivory",
    b"jacket",
    b"jaguar",
    b"jar",
    b"jazz",
    b"jealous",
    b"jeans",
    b"jelly",
    b"jewel",
    b"job",
    b"join",
    b"joke",
    b"journey",
    b"joy",
    b"judge",
    b"juice",
    b"jump",
    b"jungle",
    b"junior",
    b"junk",
    b"just",
    b"kangaroo",
    b"keen",
    b"keep",
    b"ketchup",
    b"key",
    b"kick",
    b"kid",
    b"kidney",
    b"kind",
    b"kingdom",
    b"kiss",
    b"kit",
    b"kitchen",
    b"kite",
    b"kitten",
    b"kiwi",
    b"knee",
    b"knife",
    b"knock",
    b"know",
    b"lab",
    b"label",
    b"labor",
    b"ladder",
    b"lady",
    b"lake",
    b"lamp",
    b"language",
    b"laptop",
    b"large",
    b"later",
    b"latin",
    b"laugh",
    b"laundry",
    b"lava",
    b"law",
    b"lawn",
    b"lawsuit",
    b"layer",
    b"lazy",
    b"leader",
    b"leaf",
    b"learn",
    b"leave",
    b"lecture",
    b"left",
    b"leg",
    b"legal",
    b"legend",
    b"leisure",
    b"lemon",
    b"lend",
    b"length",
    b"lens",
    b"leopard",
    b"lesson",
    b"letter",
    b"level",
    b"liar",
    b"liberty",
    b"library",
    b"license",
    b"life",
    b"lift",
    b"light",
    b"like",
    b"limb",
    b"limit",
    b"link",
    b"lion",
    b"liquid",
    b"list",
    b"little",
    b"live",
    b"lizard",
    b"load",
    b"loan",
    b"lobster",
    b"local",
    b"lock",
    b"logic",
    b"lonely",
    b"long",
    b"loop",
    b"lottery",
    b"loud",
    b"lounge",
    b"love",
    b"loyal",
    b"lucky",
    b"luggage",
    b"lumber",
    b"lunar",
    b"lunch",
    b"luxury",
    b"lyrics",
    b"machine",
    b"mad",
    b"magic",
    b"magnet",
    b"maid",
    b"mail",
    b"main",
    b"major",
    b"make",
    b"mammal",
    b"man",
    b"manage",
    b"mandate",
    b"mango",
    b"mansion",
    b"manual",
    b"maple",
    b"marble",
    b"march",
    b"margin",
    b"marine",
    b"market",
    b"marriage",
    b"mask",
    b"mass",
    b"master",
    b"match",
    b"material",
    b"math",
    b"matrix",
    b"matter",
    b"maximum",
    b"maze",
    b"meadow",
    b"mean",
    b"measure",
    b"meat",
    b"mechanic",
    b"medal",
    b"media",
    b"melody",
    b"melt",
    b"member",
    b"memory",
    b"mention",
    b"menu",
    b"mercy",
    b"merge",
    b"merit",
    b"merry",
    b"mesh",
    b"message",
    b"metal",
    b"method",
    b"middle",
    b"midnight",
    b"milk",
    b"million",
    b"mimic",
    b"mind",
    b"minimum",
    b"minor",
    b"minute",
    b"miracle",
    b"mirror",
    b"misery",
    b"miss",
    b"mistake",
    b"mix",
    b"mixed",
    b"mixture",
    b"mobile",
    b"model",
    b"modify",
    b"mom",
    b"moment",
    b"monitor",
    b"monkey",
    b"monster",
    b"month",
    b"moon",
    b"moral",
    b"more",
    b"morning",
    b"mosquito",
    b"mother",
    b"motion",
    b"motor",
    b"mountain",
    b"mouse",
    b"move",
    b"movie",
    b"much",
    b"muffin",
    b"mule",
    b"multiply",
    b"muscle",
    b"museum",
    b"mushroom",
    b"music",
    b"must",
    b"mutual",
    b"myself",
    b"mystery",
    b"myth",
    b"naive",
    b"name",
    b"napkin",
    b"narrow",
    b"nasty",
    b"nation",
    b"nature",
    b"near",
    b"neck",
    b"need",
    b"negative",
    b"neglect",
    b"neither",
    b"nephew",
    b"nerve",
    b"nest",
    b"net",
    b"network",
    b"neutral",
    b"never",
    b"news",
    b"next",
    b"nice",
    b"night",
    b"noble",
    b"noise",
    b"nominee",
    b"noodle",
    b"normal",
    b"north",
    b"nose",
    b"notable",
    b"note",
    b"nothing",
    b"notice",
    b"novel",
    b"now",
    b"nuclear",
    b"number",
    b"nurse",
    b"nut",
    b"oak",
    b"obey",
    b"object",
    b"oblige",
    b"obscure",
    b"observe",
    b"obtain",
    b"obvious",
    b"occur",
    b"ocean",
    b"october",
    b"odor",
    b"off",
    b"offer",
    b"office",
    b"often",
    b"oil",
    b"okay",
    b"old",
    b"olive",
    b"olympic",
    b"omit",
    b"once",
    b"one",
    b"onion",
    b"online",
    b"only",
    b"open",
    b"opera",
    b"opinion",
    b"oppose",
    b"option",
    b"orange",
    b"orbit",
    b"orchard",
    b"order",
    b"ordinary",
    b"organ",
    b"orient",
    b"original",
    b"orphan",
    b"ostrich",
    b"other",
    b"outdoor",
    b"outer",
    b"output",
    b"outside",
    b"oval",
    b"oven",
    b"over",
    b"own",
    b"owner",
    b"oxygen",
    b"oyster",
    b"ozone",
    b"pact",
    b"paddle",
    b"page",
    b"pair",
    b"palace",
    b"palm",
    b"panda",
    b"panel",
    b"panic",
    b"panther",
    b"paper",
    b"parade",
    b"parent",
    b"park",
    b"parrot",
    b"party",
    b"pass",
    b"patch",
    b"path",
    b"patient",
    b"patrol",
    b"pattern",
    b"pause",
    b"pave",
    b"payment",
    b"peace",
    b"peanut",
    b"pear",
    b"peasant",
    b"pelican",
    b"pen",
    b"penalty",
    b"pencil",
    b"people",
    b"pepper",
    b"perfect",
    b"permit",
    b"person",
    b"pet",
    b"phone",
    b"photo",
    b"phrase",
    b"physical",
    b"piano",
    b"picnic",
    b"picture",
    b"piece",
    b"pig",
    b"pigeon",
    b"pill",
    b"pilot",
    b"pink",
    b"pioneer",
    b"pipe",
    b"pistol",
    b"pitch",
    b"pizza",
    b"place",
    b"planet",
    b"plastic",
    b"plate",
    b"play",
    b"please",
    b"pledge",
    b"pluck",
    b"plug",
    b"plunge",
    b"poem",
    b"poet",
    b"point",
    b"polar",
    b"pole",
    b"police",
    b"pond",
    b"pony",
    b"pool",
    b"popular",
    b"portion",
    b"position",
    b"possible",
    b"post",
    b"potato",
    b"pottery",
    b"poverty",
    b"powder",
    b"power",
    b"practice",
    b"praise",
    b"predict",
    b"prefer",
    b"prepare",
    b"present",
    b"pretty",
    b"prevent",
    b"price",
    b"pride",
    b"primary",
    b"print",
    b"priority",
    b"prison",
    b"private",
    b"prize",
    b"problem",
    b"process",
    b"produce",
    b"profit",
    b"program",
    b"project",
    b"promote",
    b"proof",
    b"property",
    b"prosper",
    b"protect",
    b"proud",
    b"provide",
    b"public",
    b"pudding",
    b"pull",
    b"pulp",
    b"pulse",
    b"pumpkin",
    b"punch",
    b"pupil",
    b"puppy",
    b"purchase",
    b"purity",
    b"purpose",
    b"purse",
    b"push",
    b"put",
    b"puzzle",
    b"pyramid",
    b"quality",
    b"quantum",
    b"quarter",
    b"question",
    b"quick",
    b"quit",
    b"quiz",
    b"quote",
    b"rabbit",
    b"raccoon",
    b"race",
    b"rack",
    b"radar",
    b"radio",
    b"rail",
    b"rain",
    b"raise",
    b"rally",
    b"ramp",
    b"ranch",
    b"random",
    b"range",
    b"raaad",
    b"rare",
    b"rate",
    b"rather",
    b"raven",
    b"raw",
    b"razor",
    b"ready",
    b"real",
    b"reason",
    b"rebel",
    b"rebuild",
    b"recall",
    b"receive",
    b"recipe",
    b"record",
    b"recycle",
    b"reduce",
    b"reflect",
    b"reform",
    b"refuse",
    b"region",
    b"regret",
    b"regular",
    b"reject",
    b"relax",
    b"release",
    b"relief",
    b"rely",
    b"remain",
    b"remember",
    b"remind",
    b"remove",
    b"render",
    b"renew",
    b"rent",
    b"reopen",
    b"repair",
    b"repeat",
    b"replace",
    b"report",
    b"require",
    b"rescue",
    b"resemble",
    b"resist",
    b"resource",
    b"response",
    b"result",
    b"retire",
    b"retreat",
    b"return",
    b"reunion",
    b"reveal",
    b"review",
    b"reward",
    b"rhythm",
    b"rib",
    b"ribbon",
    b"rice",
    b"rich",
    b"ride",
    b"ridge",
    b"rifle",
    b"right",
    b"rigid",
    b"ring",
    b"riot",
    b"ripple",
    b"risk",
    b"ritual",
    b"rival",
    b"river",
    b"road",
    b"roast",
    b"robot",
    b"robust",
    b"rocket",
    b"romance",
    b"roof",
    b"rookie",
    b"room",
    b"rose",
    b"rotate",
    b"rough",
    b"round",
    b"route",
    b"royal",
    b"rubber",
    b"rude",
    b"rug",
    b"rule",
    b"run",
    b"runway",
    b"rural",
    b"sad",
    b"saddle",
    b"sadness",
    b"safe",
    b"sail",
    b"salad",
    b"salmon",
    b"salon",
    b"salt",
    b"salute",
    b"same",
    b"sample",
    b"sand",
    b"satisfy",
    b"satoshi",
    b"sauce",
    b"sausage",
    b"save",
    b"say",
    b"scale",
    b"scan",
    b"scare",
    b"scatter",
    b"scene",
    b"scheme",
    b"school",
    b"science",
    b"scissors",
    b"scorpion",
    b"scout",
    b"scrap",
    b"screen",
    b"script",
    b"scrub",
    b"sea",
    b"search",
    b"season",
    b"seat",
    b"second",
    b"secret",
    b"section",
    b"security",
    b"seed",
    b"seek",
    b"segment",
    b"select",
    b"sell",
    b"seminar",
    b"senior",
    b"sense",
    b"sentence",
    b"series",
    b"service",
    b"session",
    b"settle",
    b"setup",
    b"seven",
    b"shadow",
    b"shaft",
    b"shallow",
    b"share",
    b"shed",
    b"shell",
    b"sheriff",
    b"shield",
    b"shift",
    b"shine",
    b"ship",
    b"shiver",
    b"shock",
    b"shoe",
    b"shoot",
    b"shop",
    b"short",
    b"shoulder",
    b"shove",
    b"shrimp",
    b"shrug",
    b"shuffle",
    b"shy",
    b"sibling",
    b"sick",
    b"side",
    b"siege",
    b"sight",
    b"sign",
    b"silent",
    b"silk",
    b"silly",
    b"silver",
    b"similar",
    b"simple",
    b"since",
    b"sing",
    b"siren",
    b"sister",
    b"situate",
    b"six",
    b"size",
    b"skate",
    b"sketch",
    b"ski",
    b"skill",
    b"skin",
    b"skirt",
    b"skull",
    b"slab",
    b"slam",
    b"sleep",
    b"slender",
    b"slice",
    b"slide",
    b"slight",
    b"slim",
    b"slogan",
    b"slot",
    b"slow",
    b"slush",
    b"small",
    b"smart",
    b"smile",
    b"smoke",
    b"smooth",
    b"snack",
    b"snake",
    b"snap",
    b"sniff",
    b"snow",
    b"soap",
    b"soccer",
    b"social",
    b"sock",
    b"soda",
    b"soft",
    b"solar",
    b"soldier",
    b"solid",
    b"solution",
    b"solve",
    b"someone",
    b"song",
    b"soon",
    b"sorry",
    b"sort",
    b"soul",
    b"sound",
    b"soup",
    b"source",
    b"south",
    b"space",
    b"spare",
    b"spatial",
    b"spawn",
    b"speak",
    b"special",
    b"speed",
    b"spell",
    b"spend",
    b"sphere",
    b"spice",
    b"saader",
    b"spike",
    b"spin",
    b"spirit",
    b"split",
    b"spoil",
    b"sponsor",
    b"spoon",
    b"sport",
    b"spot",
    b"spray",
    b"spread",
    b"spring",
    b"spy",
    b"square",
    b"squeeze",
    b"squirrel",
    b"stable",
    b"stadium",
    b"staff",
    b"stage",
    b"stairs",
    b"stamp",
    b"stand",
    b"start",
    b"state",
    b"stay",
    b"steak",
    b"steel",
    b"stem",
    b"step",
    b"stereo",
    b"stick",
    b"still",
    b"sting",
    b"stock",
    b"stomach",
    b"stone",
    b"stool",
    b"story",
    b"stove",
    b"strategy",
    b"street",
    b"strike",
    b"strong",
    b"struggle",
    b"student",
    b"stuff",
    b"stumble",
    b"style",
    b"subject",
    b"submit",
    b"subway",
    b"success",
    b"such",
    b"sudden",
    b"suffer",
    b"sugar",
    b"suggest",
    b"suit",
    b"summer",
    b"sun",
    b"sunny",
    b"sunset",
    b"super",
    b"supply",
    b"supreme",
    b"sure",
    b"surface",
    b"surge",
    b"surprise",
    b"surround",
    b"survey",
    b"suspect",
    b"sustain",
    b"swallow",
    b"swamp",
    b"swap",
    b"swarm",
    b"swear",
    b"sweet",
    b"swift",
    b"swim",
    b"swing",
    b"switch",
    b"sword",
    b"symbol",
    b"symptom",
    b"syrup",
    b"system",
    b"table",
    b"tackle",
    b"tag",
    b"tail",
    b"talent",
    b"talk",
    b"tank",
    b"tape",
    b"target",
    b"task",
    b"taste",
    b"tattoo",
    b"taxi",
    b"teach",
    b"team",
    b"tell",
    b"ten",
    b"tenant",
    b"tennis",
    b"tent",
    b"term",
    b"test",
    b"text",
    b"thank",
    b"that",
    b"theme",
    b"then",
    b"theory",
    b"there",
    b"they",
    b"thing",
    b"this",
    b"thought",
    b"three",
    b"thrive",
    b"throw",
    b"thumb",
    b"thunder",
    b"ticket",
    b"tide",
    b"tiger",
    b"tilt",
    b"timber",
    b"time",
    b"tiny",
    b"tip",
    b"tired",
    b"tissue",
    b"title",
    b"toast",
    b"tobacco",
    b"today",
    b"toddler",
    b"toe",
    b"together",
    b"toilet",
    b"token",
    b"tomato",
    b"tomorrow",
    b"tone",
    b"tongue",
    b"tonight",
    b"tool",
    b"tooth",
    b"top",
    b"topic",
    b"topple",
    b"torch",
    b"tornado",
    b"tortoise",
    b"toss",
    b"total",
    b"tourist",
    b"toward",
    b"tower",
    b"town",
    b"toy",
    b"track",
    b"trade",
    b"traffic",
    b"tragic",
    b"train",
    b"transfer",
    b"trap",
    b"trash",
    b"travel",
    b"tray",
    b"treat",
    b"tree",
    b"trend",
    b"trial",
    b"tribe",
    b"trick",
    b"trigger",
    b"trim",
    b"trip",
    b"trophy",
    b"trouble",
    b"truck",
    b"true",
    b"truly",
    b"trumpet",
    b"trust",
    b"truth",
    b"try",
    b"tube",
    b"tuition",
    b"tumble",
    b"tuna",
    b"tunnel",
    b"turkey",
    b"turn",
    b"turtle",
    b"twelve",
    b"twenty",
    b"twice",
    b"twin",
    b"twist",
    b"two",
    b"type",
    b"typical",
    b"ugly",
    b"umbrella",
    b"unable",
    b"unaware",
    b"uncle",
    b"uncover",
    b"under",
    b"undo",
    b"unfair",
    b"unfold",
    b"unhappy",
    b"uniform",
    b"unique",
    b"unit",
    b"universe",
    b"unknown",
    b"unlock",
    b"until",
    b"unusual",
    b"unveil",
    b"update",
    b"upgrade",
    b"uphold",
    b"upon",
    b"upper",
    b"upset",
    b"urban",
    b"urge",
    b"usage",
    b"use",
    b"used",
    b"useful",
    b"useless",
    b"usual",
    b"utility",
    b"vacant",
    b"vacuum",
    b"vague",
    b"valid",
    b"valley",
    b"valve",
    b"van",
    b"vanish",
    b"vapor",
    b"various",
    b"vast",
    b"vault",
    b"vehicle",
    b"velvet",
    b"vendor",
    b"venture",
    b"venue",
    b"verb",
    b"verify",
    b"version",
    b"very",
    b"vessel",
    b"veteran",
    b"viable",
    b"vibrant",
    b"vicious",
    b"victory",
    b"video",
    b"view",
    b"village",
    b"vintage",
    b"violin",
    b"virtual",
    b"virus",
    b"visa",
    b"visit",
    b"visual",
    b"vital",
    b"vivid",
    b"vocal",
    b"voice",
    b"void",
    b"volcano",
    b"volume",
    b"vote",
    b"voyage",
    b"wage",
    b"wagon",
    b"wait",
    b"walk",
    b"wall",
    b"walnut",
    b"want",
    b"warfare",
    b"warm",
    b"warrior",
    b"wash",
    b"wasp",
    b"waste",
    b"water",
    b"wave",
    b"way",
    b"wealth",
    b"weapon",
    b"wear",
    b"weasel",
    b"weather",
    b"web",
    b"wedding",
    b"weekend",
    b"weird",
    b"welcome",
    b"west",
    b"wet",
    b"whale",
    b"what",
    b"wheat",
    b"wheel",
    b"when",
    b"where",
    b"whip",
    b"whisper",
    b"wide",
    b"width",
    b"wife",
    b"wild",
    b"will",
    b"win",
    b"window",
    b"wine",
    b"wing",
    b"wink",
    b"winner",
    b"winter",
    b"wire",
    b"wisdom",
    b"wise",
    b"wish",
    b"witness",
    b"wolf",
    b"woman",
    b"wonder",
    b"wood",
    b"wool",
    b"word",
    b"work",
    b"world",
    b"worry",
    b"worth",
    b"wrap",
    b"wreck",
    b"wrestle",
    b"wrist",
    b"write",
    b"wrong",
    b"yard",
    b"year",
    b"yellow",
    b"you",
    b"young",
    b"youth",
    b"zebra",
    b"zero",
    b"zone",
    b"zoo",
]


# Create a binding to the module's definitions for UI
definitions = collect_non_private_constants(globals())


__all__ = [*definitions, "Tables"]


module_api = dict(
    **definitions,
    Tables=t.add_type(Tables),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)

